#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#define MAX_FLASH_SIZE INT_MAX

struct Flash {
  int allocatedSize;
  int usedSize;
  int absoluteSize;
  int maxSize;
  unsigned char emptyChar;
  unsigned char *data;
  unsigned char *used;
} flash = {
  0, 0, 0, 0, 0xFF, NULL, NULL
};

int ExtendFlash(struct Flash *f, int bytes) {
  if (bytes > f->allocatedSize) {
    int oldAllocSize = f->allocatedSize;
    f->allocatedSize = bytes;
#if 0
    fprintf(stderr, "Alloc size %d\n", f->allocatedSize);
#endif
    if (!(f->data = realloc(f->data, f->allocatedSize*sizeof(f->data[0]))) ||
	!(f->used = realloc(f->used, f->allocatedSize*sizeof(f->used[0])))) {
      fprintf(stderr, "OUT OF MEMORY ALLOCATING FLASH ARRAY!\n");
      exit(EXIT_FAILURE);
    }
    memset(f->used+oldAllocSize, 0, f->allocatedSize-oldAllocSize);
  }
  return 0;
}

int main(int argc, char *argv[]) {
  int n, i;
  int addr = 0, filesHandled = 0;
  const char *outFileName = "-";
  FILE *outFile;

  ExtendFlash(&flash, 2);

  for (n=1; n<argc; n++) {
    if (!strcmp(argv[n], "-h")) {
      printf("CombineImg v1.11 2009-02-27 Henrik Herranen\n"
	     "Usage: %s [-h] [-o outFile] [-s n] [-f n] [+n1|file1 ...]\n"
	     , argv[0]);
      printf("-h\tHelp\n"
	     "-o file\tOutput to file (default: \"-\" = stdout)\n"
	     "-s n\tAbsolute size of image in bytes\n"
	     "-m n\tMax size of image in bytes\n"
	     "-f n\tInitially fill image with n (default 0xFF)\n"
	     "+n1\tLoad next file to address n1\n"
	     "file1\tLoad file1\n");
      exit(EXIT_SUCCESS);
    } else if (!strcmp(argv[n], "-s") || !strcmp(argv[n], "-m")) {
      if (++n >= argc) {
	fprintf(stderr, "%s: No value for parameter %s\n",
		argv[0], argv[n-1]);
	exit(EXIT_FAILURE);
      } else {
	char *tmp;
	int t = strtol(argv[n], &tmp, 0);
	if (t < 0 && *tmp) {
	  fprintf(stderr, "%s: Invalid number: %s\n", argv[0], tmp);
	  exit(EXIT_FAILURE);
	}
	if (argv[n-1][1] == 's')
	  flash.absoluteSize = t;
	else
	  flash.maxSize = t;
      }
    } else if (!strcmp(argv[n], "-o")) {
      if (++n >= argc) {
	fprintf(stderr, "%s: No value for parameter %s\n",
		argv[0], argv[n-1]);
	exit(EXIT_FAILURE);
      }
      outFileName = argv[n];
    } else if (!strcmp(argv[n], "-f")) {
      if (++n >= argc) {
	fprintf(stderr, "%s: No value for parameter %s\n",
		argv[0], argv[n-1]);
	exit(EXIT_FAILURE);
      } else if (filesHandled) {
	fprintf(stderr, "%s: Parameter %s must be before first file\n",
		argv[0], argv[n-1]);
	exit(EXIT_FAILURE);
      } else {
	char *tmp;
	long t;
	t = strtol(argv[n], &tmp, 0);
	if (t < 0 || t > 255 || *tmp) {
	  fprintf(stderr, "%s: Invalid number: %s\n", argv[0], tmp);
	  exit(EXIT_FAILURE);
	}
	flash.emptyChar = t;
      }
    } else if (argv[n][0] == '+') {
      char *tmp;
      addr = strtol(argv[n]+1, &tmp, 0);
      if (tmp && *tmp) {
	fprintf(stderr, "%s: Invalid number: %s\n", argv[0], tmp);
	exit(EXIT_FAILURE);
      }
    } else {
      FILE *fp = fopen(argv[n], "rb");
      int len;
      int startAddr = addr;
      if (!fp) {
	fprintf(stderr, "%s: Couldn't open file \"%s\"\n", argv[0], argv[n]);
	exit(EXIT_FAILURE);
      }
      fprintf(stderr, "Add file: %-20s 0x%04x..", argv[n], addr);
      while ((len = fread(flash.data+addr, 1, flash.allocatedSize-addr, fp))) {
	addr += len;
	if (addr >= flash.allocatedSize) {
	  ExtendFlash(&flash, flash.allocatedSize*3/2);
	}
      }
      fclose(fp);
      if (addr > flash.usedSize) {
	flash.usedSize = addr;
      }
      fprintf(stderr, "0x%04x\n", addr-1);
      for (i=startAddr; i<addr; i++) {
	if (flash.used[i]) {
	  fprintf(stderr, "%s: OVERLAPPING PREVIOUS AREA AT 0x%x!\n",
		  argv[0], i);
	  exit(EXIT_FAILURE);
	}
	flash.used[i] = 1;
      }
      filesHandled++;
    }
  }

#if 0
  {
    int i;
    for (i=0; i<flashSize; i++) {
      printf("%5d %3d\n", i, spi[i]);
    }
  }
#endif
  if (flash.absoluteSize && flash.usedSize > flash.absoluteSize) {
    fprintf(stderr, "%s: Used 0x%x bytes but image size is only 0x%x bytes!\n",
	    argv[0], flash.usedSize, flash.absoluteSize);
    exit(EXIT_FAILURE);
  }
  if (flash.maxSize && flash.usedSize > flash.maxSize) {
    fprintf(stderr, "%s: Used 0x%x bytes but maximum image size is only "
	    "0x%x bytes!\n",
	    argv[0], flash.usedSize, flash.maxSize);
    exit(EXIT_FAILURE);
  }
  if (flash.absoluteSize) {
    if (flash.usedSize > flash.absoluteSize) {
      ExtendFlash(&flash, flash.absoluteSize);
    }
    flash.usedSize = flash.absoluteSize;
  }
  if (!strcmp(outFileName, "-")) {
    outFile = stdout;
  } else {
    if (!(outFile = fopen(outFileName, "wb"))) {
      fprintf(stderr, "%s: Couldn't open file \"%s\" for writing!\n",
	      argv[0], outFileName);
      exit(EXIT_FAILURE);
    }
  }
  for (i=0; i<flash.usedSize; i++) {
    if (!flash.used[i])
      flash.data[i] = flash.emptyChar;
  }
  if (fwrite(flash.data, 1, flash.usedSize, outFile) < flash.usedSize) {
    fprintf(stderr, "%s: Couldn't write to file \"%s\"!\n",
	    argv[0], outFileName);
    exit(EXIT_FAILURE);
  }
  if (outFile != stdout)
    fclose(outFile);

  return EXIT_SUCCESS;
}
